@@ -1,14 +1,14 @@
-;;; inf-clojure.el --- Run an external Clojure process in an Emacs buffer -*- lexical-binding: t; -*-
+;;; inf-clojure.el --- Basic interaction with a Clojure REPL -*- lexical-binding: t; -*-
 
-;; Copyright © 2014-2022 Bozhidar Batsov
+;; Copyright © 2014-2025 Bozhidar Batsov
 
 ;; Authors: Bozhidar Batsov <bozhidar@batsov.dev>
 ;;       Olin Shivers <shivers@cs.cmu.edu>
 ;; Maintainer: Bozhidar Batsov <bozhidar@batsov.dev>
 ;; URL: http://github.com/clojure-emacs/inf-clojure
 ;; Keywords: processes, comint, clojure
-;; Version: 3.2.1
-;; Package-Requires: ((emacs "25.1") (clojure-mode "5.11"))
+;; Version: 3.3.0
+;; Package-Requires: ((emacs "28.1") (clojure-mode "5.11"))
 
 ;; This file is not part of GNU Emacs.
 
@@ -36,7 +36,7 @@
 ;; REPL.
 ;;
 ;; `inf-clojure` provides a set of essential features for interactive
-;; Clojure(Script) development:
+;; Clojure/ClojureScript/ClojureCLR development:
 ;;
 ;; * REPL
 ;; * Interactive code evaluation
@@ -47,7 +47,6 @@
 ;; * Apropos
 ;; * Macroexpansion
 ;; * Support connecting to socket REPLs
-;; * Support for Lumo
 ;; * Support for Planck
 ;; * Support for Joker
 ;;
@@ -65,19 +64,21 @@
 
 (require 'comint)
 (require 'clojure-mode)
+(require 'clojure-ts-mode nil :no-error)
 (require 'eldoc)
 (require 'thingatpt)
 (require 'ansi-color)
 (require 'cl-lib)
 (require 'subr-x)
+(require 'project)
 
 (defvar inf-clojure-startup-forms '((lein . "lein repl")
                                     (boot . "boot repl")
                                     (clojure . "clojure")
-                                    (cljs . "clojure -m cljs.main -r")
+                                    (cljs . "clojure -M -m cljs.main -r")
+                                    (lein-clr . "lein clr repl")
                                     (planck . "planck -d")
                                     (babashka . "bb")
-                                    (lumo . "lumo -d")
                                     (joker . "joker")))
 
 (defvar inf-clojure-repl-features
@@ -89,23 +90,6 @@
              (set-ns . "(in-ns '%s)")
              (macroexpand . "(cljs.core/macroexpand '%s)")
              (macroexpand-1 . "(cljs.core/macroexpand-1 '%s)")))
-    (lumo . ((load . "(clojure.core/load-file \"%s\")")
-             (doc . "(lumo.repl/doc %s)")
-             (source . "(lumo.repl/source %s)")
-             (arglists .
-                       "(let [old-value lumo.repl/*pprint-results*]
-                          (set! lumo.repl/*pprint-results* false)
-                          (js/setTimeout #(set! lumo.repl/*pprint-results* old-value) 0)
-                          (lumo.repl/get-arglists \"%s\"))")
-             (apropos . "(lumo.repl/apropos \"%s\")")
-             (ns-vars . "(lumo.repl/dir %s)")
-             (set-ns . "(in-ns '%s)")
-             (macroexpand . "(macroexpand-1 '%s)")
-             (macroexpand-1 . "(macroexpand-1 '%s)")
-             (completion .
-                         "(let [ret (atom nil)]
-                            (lumo.repl/get-completions \"%s\" (fn [res] (reset! ret (map str res))))
-                             @ret)")))
     (planck . ((load . "(load-file \"%s\")")
                (doc . "(planck.repl/doc %s)")
                (source . "(planck.repl/source %s)")
@@ -139,6 +123,17 @@
                  (set-ns . "(clojure.core/in-ns '%s)")
                  (macroexpand . "(clojure.core/macroexpand '%s)")
                  (macroexpand-1 . "(clojure.core/macroexpand-1 '%s)")))
+    (node-babashka . ((load . "(clojure.core/load-file \"%s\")")
+                 (doc . "(clojure.repl/doc %s)")
+                 (source . "(clojure.repl/source %s)")
+                 (arglists .
+                           "(try (-> '%s clojure.core/resolve clojure.core/meta :arglists)
+                              (catch Throwable e nil))")
+                 (apropos . "(doseq [var (sort (clojure.repl/apropos \"%s\"))] (println (str var)))")
+                 (ns-vars . "(clojure.repl/dir %s)")
+                 (set-ns . "(clojure.core/in-ns '%s)")
+                 (macroexpand . "(clojure.core/macroexpand '%s)")
+                 (macroexpand-1 . "(clojure.core/macroexpand-1 '%s)")))
     (clojure . ((load . "(clojure.core/load-file \"%s\")")
                 (doc . "(clojure.repl/doc %s)")
                 (source . "(clojure.repl/source %s)")
@@ -153,7 +148,22 @@
                 (ns-vars . "(clojure.repl/dir %s)")
                 (set-ns . "(clojure.core/in-ns '%s)")
                 (macroexpand . "(clojure.core/macroexpand '%s)")
-                (macroexpand-1 . "(clojure.core/macroexpand-1 '%s)")))))
+                (macroexpand-1 . "(clojure.core/macroexpand-1 '%s)")))
+    (lein-clr . ((load . "(clojure.core/load-file \"%s\")")
+                 (doc . "(clojure.repl/doc %s)")
+                 (source . "(clojure.repl/source %s)")
+                 (arglists .
+                           "(try
+                             (:arglists
+                              (clojure.core/meta
+                               (clojure.core/resolve
+                                (clojure.core/read-string \"%s\"))))
+                             (catch Exception e nil))")
+                 (apropos . "(doseq [var (sort (clojure.repl/apropos \"%s\"))] (println (str var)))")
+                 (ns-vars . "(clojure.repl/dir %s)")
+                 (set-ns . "(clojure.core/in-ns '%s)")
+                 (macroexpand . "(clojure.core/macroexpand '%s)")
+                 (macroexpand-1 . "(clojure.core/macroexpand-1 '%s)")))))
 
 (defvar-local inf-clojure-repl-type nil
   "Symbol to define your REPL type.
@@ -166,10 +176,10 @@ either `setq-local` or an entry in `.dir-locals.el`." )
 MULTIPLE PROCESS SUPPORT
 ===========================================================================
 To run multiple Clojure processes, you start the first up
-with \\[inf-clojure].  It will be in a buffer named `*inf-clojure*'.
+with \\[inf-clojure].  It will be in a buffer named *inf-clojure*.
 Rename this buffer with \\[rename-buffer].  You may now start up a new
 process with another \\[inf-clojure].  It will be in a new buffer,
-named `*inf-clojure*'.  You can switch between the different process
+named *inf-clojure*.  You can switch between the different process
 buffers with \\[switch-to-buffer].
 
 Commands that send text from source buffers to Clojure processes --
@@ -178,7 +188,7 @@ process to send to, when you have more than one Clojure process around.  This
 is determined by the global variable `inf-clojure-buffer'.  Suppose you
 have three inferior Clojures running:
     Buffer              Process
-    foo                 inf-clojure
+    foo                 `inf-clojure'
     bar                 inf-clojure<2>
     *inf-clojure*     inf-clojure<3>
 If you do a \\[inf-clojure-eval-defun] command on some Clojure source code,
@@ -213,7 +223,7 @@ If no-error is truthy don't error if feature is not present."
 
 (defun inf-clojure--update-feature (repl-type feature form)
   "Return a copy of the datastructure containing the repl features.
-Given a REPL-TYPE (`clojure', `lumo', ...) and a FEATURE (`doc',
+Given a REPL-TYPE (`clojure', `planck', ...) and a FEATURE (`doc',
 `apropos', ...) and a FORM this will return a new datastructure
 that can be set as `inf-clojure-repl-features'."
   (let ((original (alist-get repl-type inf-clojure-repl-features)))
@@ -226,7 +236,7 @@ that can be set as `inf-clojure-repl-features'."
 
 (defun inf-clojure-update-feature (repl-type feature form)
   "Mutate the repl features to the new FORM.
-Given a REPL-TYPE (`clojure', `lumo', ...) and a FEATURE (`doc',
+Given a REPL-TYPE (`clojure', `planck', ...) and a FEATURE (`doc',
 `apropos', ...) and a FORM this will set
 `inf-clojure-repl-features' with these new values."
   (setq inf-clojure-repl-features (inf-clojure--update-feature repl-type feature form)))
@@ -242,7 +252,7 @@ has been found.  See also variable `inf-clojure-buffer'."
         (error "No Clojure subprocess; see variable `inf-clojure-buffer'"))))
 
 (defun inf-clojure-repl-p (&optional buf)
-  "Indicates if BUF is an inf-clojure REPL.
+  "Indicates if BUF is an `inf-clojure' REPL.
 If BUF is nil then defaults to the current buffer.
 Checks the mode and that there is a live process."
   (let ((buf (or buf (current-buffer))))
@@ -251,28 +261,37 @@ Checks the mode and that there is a live process."
          (process-live-p (get-buffer-process buf)))))
 
 (defun inf-clojure-repls ()
-  "Return a list of all inf-clojure REPL buffers."
+  "Return a list of all `inf-clojure' REPL buffers."
   (let (repl-buffers)
     (dolist (b (buffer-list))
       (when (inf-clojure-repl-p b)
         (push (buffer-name b) repl-buffers)))
     repl-buffers))
 
+(defun inf-clojure--prompt-repl-buffer (prompt)
+  "Prompt the user to select an `inf-clojure' repl buffer.
+PROMPT is a string to prompt the user.
+Returns nil when no buffer is selected."
+  (let ((repl-buffers (inf-clojure-repls)))
+    (if (> (length repl-buffers) 0)
+        (when-let* ((repl-buffer (completing-read prompt repl-buffers nil t)))
+          (get-buffer repl-buffer))
+      (user-error "No buffers have an inf-clojure process"))))
+
 (defun inf-clojure-set-repl (always-ask)
-  "Set an inf-clojure buffer as the active (default) REPL.
+  "Set an `inf-clojure' buffer as the active (default) REPL.
 If in a REPL buffer already, use that unless a prefix is used (or
-ALWAYS-ASK).  Otherwise get a list of all active inf-clojure
+ALWAYS-ASK).  Otherwise get a list of all active `inf-clojure'
 REPLS and offer a choice.  It's recommended to rename REPL
 buffers after they are created with `rename-buffer'."
   (interactive "P")
-  (if (and (not always-ask)
-           (inf-clojure-repl-p))
-      (setq inf-clojure-buffer (current-buffer))
-    (let ((repl-buffers (inf-clojure-repls)))
-     (if (> (length repl-buffers) 0)
-         (when-let ((repl-buffer (completing-read "Select default REPL: " repl-buffers nil t)))
-           (setq inf-clojure-buffer (get-buffer repl-buffer)))
-       (user-error "No buffers have an inf-clojure process")))))
+  (when-let* ((new-repl-buffer
+               (if (or always-ask
+                       (not (inf-clojure-repl-p)))
+                   (inf-clojure--prompt-repl-buffer "Select default REPL: ")
+                 (current-buffer))))
+    (setq inf-clojure-buffer new-repl-buffer)
+    (message "Current inf-clojure REPL set to %s" new-repl-buffer)))
 
 (defvar inf-clojure--repl-type-lock nil
   "Global lock for protecting against proc filter race conditions.
@@ -294,9 +313,7 @@ See http://blog.jorgenschaefer.de/2014/05/race-conditions-in-emacs-process-filte
   :link '(emacs-commentary-link :tag "Commentary" "inf-clojure"))
 
 (defconst inf-clojure-version
-  (or (if (fboundp 'package-get-version)
-          (package-get-version))
-      "3.2.1")
+  (package-get-version)
   "The current version of `inf-clojure'.")
 
 (defcustom inf-clojure-prompt-read-only t
@@ -313,6 +330,14 @@ mode.  Default is whitespace followed by 0 or 1 single-letter colon-keyword
 \(as in :a, :c, etc.)"
   :type 'regexp)
 
+(defcustom inf-clojure-source-modes '(clojure-ts-mode clojure-mode)
+  "Used to determine if a buffer contains Clojure source code.
+
+Any buffer with one of these major modes, it's considered a Clojure
+source file by all `inf-clojure' commands."
+  :type '(repeat symbol)
+  :safe #'symbolp)
+
 (defun inf-clojure--modeline-info ()
   "Return modeline info for `inf-clojure-minor-mode'.
 Either \"no process\" or \"buffer-name(repl-type)\""
@@ -323,7 +348,8 @@ Either \"no process\" or \"buffer-name(repl-type)\""
     "no process"))
 
 (defvar inf-clojure-mode-map
-  (let ((map (copy-keymap comint-mode-map)))
+  (let ((map (make-sparse-keymap)))
+    (set-keymap-parent map comint-mode-map)
     (define-key map (kbd "C-x C-e") #'inf-clojure-eval-last-sexp)
     (define-key map (kbd "C-c C-l") #'inf-clojure-load-file)
     (define-key map (kbd "C-c C-a") #'inf-clojure-show-arglists)
@@ -416,7 +442,7 @@ The value of this variable is a mode line template as in
 `mode-line-format'.  See Info Node `(elisp)Mode Line Format' for details
 about mode line templates.
 
-Customize this variable to change how inf-clojure-minor-mode
+Customize this variable to change how `inf-clojure-minor-mode'
 displays its status in the mode line.  The default value displays
 the current REPL.  Set this variable to nil to disable the
 mode line entirely."
@@ -475,12 +501,14 @@ Should be a symbol that is a key in `inf-clojure-repl-features'."
   :package-version '(inf-clojure . "3.0.0")
   :type '(choice (const :tag "clojure" clojure)
                  (const :tag "cljs" cljs)
-                 (const :tag "lumo" lumo)
                  (const :tag "planck" planck)
                  (const :tag "joker" joker)
                  (const :tag "babashka" babashka)
                  (const :tag "determine at startup" nil)))
 
+(defvar inf-clojure-custom-repl-name nil
+  "A string to be used as the repl buffer name.")
+
 (defun inf-clojure--whole-comment-line-p (string)
   "Return non-nil iff STRING is a whole line semicolon comment."
   (string-match-p "^\s*;" string))
@@ -569,22 +597,35 @@ This should usually be a combination of `inf-clojure-prompt' and
   :package-version '(inf-clojure . "2.0.0"))
 
 (defcustom inf-clojure-auto-mode t
-  "When non-nil, automatically enable inf-clojure-minor-mode for all Clojure buffers."
+  "Automatically enable `inf-clojure-minor-mode'.
+All buffers in `clojure-mode' will automatically be in
+`inf-clojure-minor-mode' unless set to nil."
   :type 'boolean
   :safe #'booleanp
   :package-version '(inf-clojure . "3.1.0"))
 
+(defun inf-clojure--get-preferred-major-modes ()
+  "Return list of preferred major modes that are actually available."
+  (cl-remove-if-not (lambda (mode) (featurep mode))
+                    inf-clojure-source-modes))
+
+(defun inf-clojure--clojure-buffer-p ()
+  "Return TRUE if the current buffer is a Clojure buffer."
+  (derived-mode-p (inf-clojure--get-preferred-major-modes)))
+
 (defun inf-clojure--clojure-buffers ()
   "Return a list of all existing `clojure-mode' buffers."
-  (cl-remove-if-not
-   (lambda (buffer) (with-current-buffer buffer (derived-mode-p 'clojure-mode)))
-   (buffer-list)))
+  (cl-remove-if-not (lambda (buffer)
+                      (with-current-buffer buffer
+                        (inf-clojure--clojure-buffer-p)))
+                    (buffer-list)))
 
 (defun inf-clojure-enable-on-existing-clojure-buffers ()
   "Enable inf-clojure's minor mode on existing Clojure buffers.
 See command `inf-clojure-minor-mode'."
   (interactive)
-  (add-hook 'clojure-mode-hook #'inf-clojure-minor-mode)
+  (dolist (mode (inf-clojure--get-preferred-major-modes))
+    (add-hook (derived-mode-hook-name mode) #'inf-clojure-minor-mode))
   (dolist (buffer (inf-clojure--clojure-buffers))
     (with-current-buffer buffer
       (inf-clojure-minor-mode +1))))
@@ -615,36 +656,39 @@ Customization: Entry to this mode runs the hooks on `comint-mode-hook' and
 
 You can send text to the inferior Clojure process from other buffers containing
 Clojure source.
-    `inf-clojure-switch-to-repl' switches the current buffer to the Clojure process buffer.
+    `inf-clojure-switch-to-repl' switches the current buffer to the Clojure
+    process buffer.
     `inf-clojure-eval-defun' sends the current defun to the Clojure process.
     `inf-clojure-eval-region' sends the current region to the Clojure process.
 
     Prefixing the inf-clojure-eval/defun/region commands with
-    a \\[universal-argument] causes a switch to the Clojure process buffer after sending
-    the text.
+    a \\[universal-argument] causes a switch to the Clojure process buffer after
+    sending the text.
 
 Commands:\\<inf-clojure-mode-map>
-\\[comint-send-input] after the end of the process' output sends the text from the
-    end of process to point.
-\\[comint-send-input] before the end of the process' output copies the sexp ending at point
-    to the end of the process' output, and sends it.
-\\[comint-copy-old-input] copies the sexp ending at point to the end of the process' output,
-    allowing you to edit it before sending it.
-If `comint-use-prompt-regexp' is nil (the default), \\[comint-insert-input] on old input
-   copies the entire old input to the end of the process' output, allowing
-   you to edit it before sending it.  When not used on old input, or if
-   `comint-use-prompt-regexp' is non-nil, \\[comint-insert-input] behaves according to
-   its global binding.
+\\[comint-send-input] after the end of the process' output sends the text from
+    the end of process to point.
+\\[comint-send-input] before the end of the process' output copies the sexp
+    ending at point to the end of the process' output, and sends it.
+\\[comint-copy-old-input] copies the sexp ending at point to the end of the
+    process' output,allowing you to edit it before sending it.
+If `comint-use-prompt-regexp' is nil (the default), \\[comint-insert-input] on
+   old input copies the entire old input to the end of the process' output,
+   allowing you to edit it before sending it.  When not used on old input, or if
+   `comint-use-prompt-regexp' is non-nil, \\[comint-insert-input] behaves
+   according to its global binding.
 \\[backward-delete-char-untabify] converts tabs to spaces as it moves back.
 \\[clojure-indent-line] indents for Clojure; with argument, shifts rest
     of expression rigidly with the current line.
-\\[indent-sexp] does \\[clojure-indent-line] on each line starting within following expression.
-Paragraphs are separated only by blank lines.  Semicolons start comments.
-If you accidentally suspend your process, use \\[comint-continue-subjob]
-to continue it."
+\\[indent-sexp] does \\[clojure-indent-line] on each line starting within
+  following expression.  Paragraphs are separated only by blank lines.
+  Semicolons start comments.  If you accidentally suspend your process,
+  use \\[comint-continue-subjob] to continue it."
   (setq comint-input-sender 'inf-clojure--send-string)
   (setq comint-prompt-regexp inf-clojure-comint-prompt-regexp)
   (setq mode-line-process '(":%s"))
+  ;; NOTE: Using Tree-sitter based syntax highlighting in comint
+  ;; buffer is currently not possible.
   (clojure-mode-variables)
   (clojure-font-lock-setup)
   (when inf-clojure-enable-eldoc
@@ -756,21 +800,39 @@ to suppress the usage of the target buffer discovery logic."
 The name is simply the final segment of the path."
   (file-name-nondirectory (directory-file-name dir)))
 
+(defun inf-clojure--project-dir ()
+  "Return current Clojure project root."
+  (let ((project-vc-extra-root-markers '("project.clj"      ; Leiningen
+                                         "build.gradle"     ; Gradle
+                                         "build.gradle.kts" ; Gradle
+                                         "deps.edn"         ; Clojure CLI (a.k.a. tools.deps)
+                                         "shadow-cljs.edn"  ; shadow-cljs
+                                         "bb.edn"           ; babashka
+                                         "nbb.edn"          ; nbb
+                                         "basilisp.edn"     ; Basilisp (Python)
+                                         )))
+    (project-root (project-current))))
+
+(defvar clojure-ts-mode-syntax-table)
+
 ;;;###autoload
-(defun inf-clojure (cmd)
-  "Run an inferior Clojure process, input and output via buffer `*inf-clojure*'.
-If there is a process already running in `*inf-clojure*', just
+(defun inf-clojure (cmd &optional suppress-message)
+  "Run an inferior Clojure process, input and output via buffer *inf-clojure*.
+If there is a process already running in *inf-clojure*, just
 switch to that buffer.
 
 CMD is a string which serves as the startup command or a cons of
 host and port.
 
- Prompts user for repl startup command and repl type if not
+ Prompts user for REPL startup command and REPL type if not
 inferrable from startup command.  Uses `inf-clojure-custom-repl-type'
 and `inf-clojure-custom-startup' if those are set.
 Use a prefix to prevent using these when they
 are set.
 
+Prints a message that it has connected to the host and port
+unless SUPPRESS-MESSAGE is truthy.
+
  Runs the hooks from `inf-clojure-mode-hook' (after the
 `comint-mode-hook' is run).  \(Type \\[describe-mode] in the
 process buffer for a list of commands.)"
@@ -780,10 +842,12 @@ process buffer for a list of commands.)"
                                           (mapcar #'cdr inf-clojure-startup-forms)
                                           nil
                                           'confirm-after-completion))))
-  (let* ((project-dir (clojure-project-dir))
-         (process-buffer-name (if project-dir
-                                  (format "inf-clojure %s" (inf-clojure--project-name project-dir))
-                                "inf-clojure"))
+  (let* ((project-dir (inf-clojure--project-dir))
+         (process-buffer-name (or
+                               inf-clojure-custom-repl-name
+                               (if project-dir
+                                   (format "inf-clojure %s" (inf-clojure--project-name project-dir))
+                                 "inf-clojure")))
          ;; comint adds the asterisks to both sides
          (repl-buffer-name (format "*%s*" process-buffer-name)))
     ;; Create a new comint buffer if needed
@@ -797,25 +861,131 @@ process buffer for a list of commands.)"
                              inf-clojure-custom-repl-type)
                            (car (rassoc cmd inf-clojure-startup-forms))
                            (inf-clojure--prompt-repl-type))))
-        (message "Starting Clojure REPL via `%s'..." cmd)
+        (unless suppress-message
+          (message "Starting Clojure REPL via `%s'..." cmd))
         (with-current-buffer (apply #'make-comint
                                     process-buffer-name (car cmdlist) nil (cdr cmdlist))
           (inf-clojure-mode)
-          (set-syntax-table clojure-mode-syntax-table)
+          (set-syntax-table (pcase (car (inf-clojure--get-preferred-major-modes))
+                              ('clojure-ts-mode clojure-ts-mode-syntax-table)
+                              (_ clojure-mode-syntax-table)))
           (setq-local inf-clojure-repl-type repl-type)
           (hack-dir-local-variables-non-file-buffer))))
     ;; update the default comint buffer and switch to it
     (setq inf-clojure-buffer (get-buffer repl-buffer-name))
     (if inf-clojure-repl-use-same-window
         (pop-to-buffer-same-window repl-buffer-name)
-      (pop-to-buffer repl-buffer-name))))
+      (pop-to-buffer repl-buffer-name))
+    repl-buffer-name))
+
+;;;###autol
+(defun inf-clojure-connect (host port &optional suppress-message)
+  "Connect to a running socket REPL server via `inf-clojure'.
+HOST is the host the process is running on, PORT is where it's
+listening.  SUPPRESS-MESSAGE is optional and if truthy will
+prevent showing the startup message."
+  (interactive "shost: \nnport: ")
+  (inf-clojure (cons host port) suppress-message))
+
+(defvar-local inf-clojure-socket-callback nil
+  "Used to transfer state between the socket process buffer & REPL buffer.")
+
+(defvar-local inf-clojure-socket-buffer nil
+  "Used to kill the associated socket buffer when it's REPL buffer is killed.")
+
+(defun inf-clojure-socket-filter (process output)
+  "A filter that gets triggered each time the socket receives new OUTPUT.
+This function prints out the output received but also
+watches for a prompt using the `inf-clojure-prompt' regexp, once
+this happens a callback is triggered if available.  The callback
+is intended to be used to trigger a `inf-clojure-connect' once we
+can determine that a socket REPL is ready to receive a
+connection.
+
+PROCESS is the process object that is being filtered.
+
+OUTPUT is the latest data received from the process"
+  (let ((server-buffer (process-buffer process)))
+    (when (buffer-live-p server-buffer)
+      (with-current-buffer server-buffer
+        (insert output)))
+    (let ((prompt-displayed (string-match inf-clojure-prompt output)))
+      (when prompt-displayed
+	(with-current-buffer server-buffer
+	  (when inf-clojure-socket-callback
+	    (funcall inf-clojure-socket-callback)))))))
+
+(defun inf-clojure-socket-repl-sentinel (process _event)
+  "Ensures socket REPL are cleaned up when the REPL buffer is closed.
+
+PROCESS is the process object that is connected to a socket REPL.
+
+EVENT is the event that triggered this function to be called."
+  (when (not (process-live-p process))
+    (let ((repl-buffer (process-buffer process)))
+      (with-current-buffer repl-buffer
+        (when inf-clojure-socket-buffer
+          (kill-buffer inf-clojure-socket-buffer))))))
+
+(defvar inf-clojure-socket-repl-startup-forms
+  '((lein . "JVM_OPTS='-Dclojure.server.repl={:port %d :accept clojure.core.server/repl}' lein repl")
+    (boot . "export BOOT_JVM_OPTIONS='-Dclojure.server.repl=\"{:port %d :accept clojure.core.server/repl}\"' boot repl")
+    (clojure . "clojure -J-Dclojure.server.repl=\"{:port %d :accept clojure.core.server/repl}\"")
+    (cljs . "clojure -J-Dclojure.server.repl=\"{:port %d :accept cljs.server.browser/repl}\"")
+    (lein-clr . "JVM_OPTS='-Dclojure.server.repl={:port %d :accept clojure.core.server/repl}' lein clr repl")
+    (planck . "planck -n %d")
+    (babashka . "bb socket-repl %d")))
+
+(defcustom inf-clojure-socket-repl-port
+  nil
+  "Port to be used when creating a socket REPL via `inf-clojure-socket-repl'.
+If left as nil a random port will be selected between 5500-6000."
+  :type '(choice integer (const nil))
+  :package-version '(inf-clojure . "3.3"))
 
 ;;;###autoload
-(defun inf-clojure-connect (host port)
-  "Connect to a running socket REPL server via `inf-clojure'.
-HOST is the host the process is running on, PORT is where it's listening."
-  (interactive "shost: \nnport: ")
-  (inf-clojure (cons host port)))
+(defun inf-clojure-socket-repl (cmd)
+  "Start a socket REPL server and connects to it via `inf-clojure-connect'.
+CMD is the command line instruction used to start the socket
+REPL.  It should be a string with \"%d\" in it to take a random
+port.  Set `inf-clojure-custom-startup' or choose from the
+defaults provided in `inf-clojure-socket-repl-startup-forms'."
+  (interactive (list (or (unless current-prefix-arg
+                           inf-clojure-custom-startup)
+                         (completing-read "Select Clojure socket REPL startup command: "
+                                          (mapcar #'cdr inf-clojure-socket-repl-startup-forms)
+                                          nil
+                                          'confirm-after-completion))))
+  (let* ((host "localhost")
+         (port (or inf-clojure-socket-repl-port (+ 5500 (random 500))))
+         (project-dir (inf-clojure--project-dir))
+         (repl-type (or (unless prefix-arg
+                          inf-clojure-custom-repl-type)
+                        (car (rassoc cmd inf-clojure-socket-repl-startup-forms))
+                        (inf-clojure--prompt-repl-type)))
+         (project-name (inf-clojure--project-name (or project-dir "standalone")))
+         (socket-process-name (format "*%s-%s-socket-server*" project-name repl-type))
+         (socket-buffer (get-buffer-create
+                         (format "*%s-%s-socket*" project-name repl-type)))
+         (socket-cmd (format cmd port))
+         (sock (let ((default-directory (or project-dir default-directory)))
+                 (start-file-process-shell-command
+                  socket-process-name socket-buffer
+                  socket-cmd))))
+    (with-current-buffer socket-buffer
+      (setq-local
+       inf-clojure-socket-callback
+       (lambda ()
+         (let* ((inf-clojure-custom-repl-type repl-type)
+                (created-repl-buffer (inf-clojure-connect host port :suppress-message)))
+           (with-current-buffer (get-buffer created-repl-buffer)
+             (setq-local inf-clojure-socket-buffer socket-buffer)
+             (set-process-sentinel
+              (get-buffer-process (current-buffer))
+              #'inf-clojure-socket-repl-sentinel))))))
+    (set-process-filter sock #'inf-clojure-socket-filter)
+    (message "Starting %s socket REPL server at %s:%d with %s" repl-type host port socket-cmd)))
+
 
 (defun inf-clojure--forms-without-newlines (str)
   "Remove newlines between toplevel forms.
@@ -826,17 +996,24 @@ of forms."
   (condition-case nil
       (with-temp-buffer
         (progn
-          (clojurec-mode)
+          ;; Activate preferred major mode.
+          (funcall (car (inf-clojure--get-preferred-major-modes)))
           (insert str)
           (whitespace-cleanup)
           (goto-char (point-min))
           (while (not (eobp))
             (while (looking-at "\n")
               (delete-char 1))
+            ;; NOTE: There is no special API for that in
+            ;; `clojure-ts-mode', so probably for now lets keep this
+            ;; `clojure-mode' function.
             (unless (eobp)
               (clojure-forward-logical-sexp))
             (unless (eobp)
-              (forward-char)))
+              (forward-char)
+              ;; Remove an empty line at the end of the buffer.
+              (when (eobp)
+                (delete-char -1))))
           (buffer-substring-no-properties (point-min) (point-max))))
     (scan-error str)))
 
@@ -866,7 +1043,7 @@ current defun else return the string.."
     (let ((end (point))
           (case-fold-search t)
           (func (if bounds #'cons #'buffer-substring-no-properties)))
-      (beginning-of-defun)
+      (beginning-of-defun-raw)
       (funcall func (point) end))))
 
 (defun inf-clojure-eval-defun (&optional and-go)
@@ -904,16 +1081,21 @@ Prefix argument AND-GO means switch to the Clojure buffer afterwards."
   "Insert FORM into process and evaluate.
 Indent FORM.  FORM is expected to have been trimmed."
   (let ((clojure-process (inf-clojure-proc)))
-    (with-current-buffer (process-buffer clojure-process)
-      (comint-goto-process-mark)
-      (let ((beginning (point)))
-        (insert (format "%s" form))
-        (let ((end (point)))
-          (goto-char beginning)
-          (indent-sexp end)
-          ;; font-lock the inserted code
-          (font-lock-ensure beginning end)))
-      (comint-send-input t t))))
+    ;; ensure the repl buffer scrolls. See similar fix in CIDER:
+    ;; https://github.com/clojure-emacs/cider/pull/2590
+    (with-selected-window (or (get-buffer-window inf-clojure-buffer)
+                              (selected-window))
+      (with-current-buffer (process-buffer clojure-process)
+        (comint-goto-process-mark)
+        (let ((beginning (point)))
+          (insert form)
+          (let ((end (point)))
+            (goto-char beginning)
+            (indent-sexp end)
+            ;; font-lock the inserted code
+            (font-lock-ensure beginning end)
+            (goto-char end)))
+        (comint-send-input t t)))))
 
 (defun inf-clojure-insert-defun ()
   "Send current defun to process."
@@ -943,18 +1125,11 @@ START and END are the beginning and end positions in the buffer to send."
   (interactive)
   (inf-clojure-eval-defun t))
 
-(defvar inf-clojure-prev-l/c-dir/file nil
+(defvar inf-clojure-prev-loaded-dir-and-file nil
   "Record last directory and file used in loading or compiling.
 This holds a cons cell of the form `(DIRECTORY . FILE)'
 describing the last `inf-clojure-load-file' command.")
 
-(defcustom inf-clojure-source-modes '(clojure-mode)
-  "Used to determine if a buffer contains Clojure source code.
-If it's loaded into a buffer that is in one of these major modes, it's
-considered a Clojure source file by `inf-clojure-load-file'.
-Used by this command to determine defaults."
-  :type '(repeat symbol))
-
 (defun inf-clojure-load-file (&optional switch-to-repl file-name)
   "Load a Clojure file into the inferior Clojure process.
 
@@ -964,34 +1139,45 @@ is present it will be used instead of the current file."
   (interactive "P")
   (let* ((proc (inf-clojure-proc))
          (file-name (or file-name
-                        (car (comint-get-source "Load Clojure file: " inf-clojure-prev-l/c-dir/file
+                        (car (comint-get-source "Load Clojure file: " inf-clojure-prev-loaded-dir-and-file
                                                 ;; nil because doesn't need an exact name
-                                                inf-clojure-source-modes nil))))
+                                                (inf-clojure--get-preferred-major-modes) nil))))
          (load-form (inf-clojure-get-feature proc 'load)))
     (comint-check-source file-name) ; Check to see if buffer needs saved.
-    (setq inf-clojure-prev-l/c-dir/file (cons (file-name-directory    file-name)
+    (setq inf-clojure-prev-loaded-dir-and-file (cons (file-name-directory    file-name)
                                               (file-name-nondirectory file-name)))
     (inf-clojure--send-string proc (format load-form file-name))
     (when switch-to-repl
       (inf-clojure-switch-to-repl t))))
 
+(declare-function clojure-ts-find-ns "clojure-ts-mode")
+
+(defun inf-clojure--find-ns ()
+  "Return the namespace of the current Clojure buffer.
+
+This function delegates its job to an appropritate function, considering
+`inf-clojure-source-modes'."
+  (pcase (car (inf-clojure--get-preferred-major-modes))
+    ('clojure-ts-mode (clojure-ts-find-ns))
+    (_ (clojure-find-ns))))
+
 (defun inf-clojure-reload (arg)
   "Send a query to the inferior Clojure for reloading the namespace.
-See variable `inf-clojure-reload-form' and
+See variable `inf-clojure-reload-form' and variable
 `inf-clojure-reload-all-form'.
 
 The prefix argument ARG can change the behavior of the command:
 
-  - C-u M-x `inf-clojure-reload': prompts for a namespace name.
-  - M-- M-x `inf-clojure-reload': executes (require ... :reload-all).
-  - M-- C-u M-x `inf-clojure-reload': reloads all AND prompts."
+  - \\`C-u' \\[inf-clojure-reload]: prompts for a namespace name.
+  - \\`M--' \\[inf-clojure-reload]: executes (require ... :reload-all).
+  - \\`M--' \\`C-u' \\[inf-clojure-reload]: reloads all AND prompts."
   (interactive "P")
   (let* ((proc (inf-clojure-proc))
          (reload-all-p (or (equal arg '-) (equal arg '(-4))))
          (prompt-p (or (equal arg '(4)) (equal arg '(-4))))
          (ns (if prompt-p
-                (car (inf-clojure-symprompt "Namespace" (clojure-find-ns)))
-              (clojure-find-ns)))
+                (car (inf-clojure-symprompt "Namespace" (inf-clojure--find-ns)))
+              (inf-clojure--find-ns)))
          (form (if (not reload-all-p)
                    (inf-clojure-reload-form proc)
                  (inf-clojure-reload-all-form proc))))
@@ -1092,7 +1278,7 @@ STRING if present."
                             (prin1-to-string (substring-no-properties string))))
                   nil
                   (expand-file-name inf-clojure--log-file-name
-                                    (clojure-project-dir))
+                                    (inf-clojure--project-dir))
                   'append
                   'no-annoying-write-file-in-minibuffer)))
 
@@ -1200,11 +1386,11 @@ for evaluation, therefore FORM should not include it."
 (defun inf-clojure-arglists (fn)
   "Send a query to the inferior Clojure for the arglists for function FN.
 See variable `inf-clojure-arglists-form'."
-  (when-let ((proc (inf-clojure-proc 'no-error)))
-    (when-let ((arglists-form (inf-clojure-get-feature proc 'arglists)))
-      (thread-first (format arglists-form fn)
-        (inf-clojure--process-response proc "(" ")")
-        (inf-clojure--some)))))
+  (when-let* ((proc (inf-clojure-proc 'no-error))
+              (arglists-form (inf-clojure-get-feature proc 'arglists)))
+    (thread-first (format arglists-form fn)
+                  (inf-clojure--process-response proc "(" ")")
+                  (inf-clojure--some))))
 
 (defun inf-clojure-show-arglists (prompt-for-symbol)
   "Show the arglists for function FN in the mini-buffer.
@@ -1226,8 +1412,8 @@ prefix argument PROMPT-FOR-NS, it prompts for a namespace name."
   (interactive "P")
   (let* ((proc (inf-clojure-proc))
          (ns (if prompt-for-ns
-                 (car (inf-clojure-symprompt "Ns vars" (clojure-find-ns)))
-               (clojure-find-ns)))
+                 (car (inf-clojure-symprompt "Ns vars" (inf-clojure--find-ns)))
+               (inf-clojure--find-ns)))
          (ns-vars-form (inf-clojure-get-feature proc 'ns-vars)))
     (inf-clojure--send-string proc (format ns-vars-form ns))))
 
@@ -1239,8 +1425,8 @@ PROMPT-FOR-NS, it prompts for a namespace name."
   (interactive "P")
   (let* ((proc (inf-clojure-proc))
          (ns (if prompt-for-ns
-                 (car (inf-clojure-symprompt "Set ns to" (clojure-find-ns)))
-               (clojure-find-ns)))
+                 (car (inf-clojure-symprompt "Set ns to" (inf-clojure--find-ns)))
+               (inf-clojure--find-ns)))
          (set-ns-form (inf-clojure-get-feature proc 'set-ns)))
     (when (or (not ns) (equal ns ""))
       (user-error "No namespace selected"))
@@ -1321,22 +1507,13 @@ evaluating \\[inf-clojure-completion-form] at the REPL."
         (funcall inf-clojure-completions-fn
                  (inf-clojure--process-response completion-expr proc  "(" ")"))))))
 
-(defconst inf-clojure-clojure-expr-break-chars "^[] \"'`><,;|&{()[@\\^]"
-  "Regexp are hard.
+(defconst inf-clojure-clojure-expr-break-chars "^][ \"'`><,;|&{()@\\^"
+  "A list of characters that serve as expression boundaries.
 
-This regex has been built in order to match the first of the
-listed chars.  There are a couple of quirks to consider:
-
-- the ] is always a special in elisp regex so you have to put it
-  directly AFTER [ if you want to match it as literal.
-- The ^ needs to be escaped with \\^.
-
-Tests and `re-builder' are your friends.")
+See `inf-clojure-completion-bounds-of-expr-at-point'.")
 
 (defun inf-clojure--kw-to-symbol (kw)
-  "Convert the keyword KW to a symbol.
-
-This guy was taken from CIDER, thanks folks."
+  "Convert the keyword KW to a symbol."
   (when kw
     (replace-regexp-in-string "\\`:+" "" kw)))
 
